pub mod widgets;
use crate::gui::widgets::WidgetType::About;
use egui::{Ui, menu, Color32, FontDefinitions};
use egui_tiles::{Container, SimplificationOptions, Tile, TileId, Tiles};
use log::{error, warn, info, debug, trace};
use widgets::about;

/// Runs the actual GUI.
/// We contruct the empty tree, headers and basic window stuff here.
/// Widgets are all opened via this, or loaded from a saved state.
pub(crate) async fn run_gui() -> Result<(), eframe::Error> {
    // Start Logging
    egui_logger::builder().init().expect("Error initialising egui_logger");
    debug!("Running in GUI Mode!");
    // run actual ui
    let options = eframe::NativeOptions {
        viewport: egui::ViewportBuilder::default().with_inner_size([1920.0, 1080.0]),
        ..Default::default()
    };
    eframe::run_native(
        "Y(et) A(nother) M(eshtastic) S(erver) - v0.1.0",
        options,
        Box::new(|_cc| {
            #[cfg_attr(not(feature = "serde"), allow(unused_mut))]
            let mut app = YamsApp::default();
            #[cfg(feature = "serde")]
            if let Some(storage) = _cc.storage {
                if let Some(state) = eframe::get_value(storage, eframe::APP_KEY) {
                    app = state;
                }
            }
            Ok(Box::new(app))
        }),
    )
}

/// This allows us to add new widgets into the UI.
impl YamsApp {
    // Add a new pane to the tree, depending on the widget_type.
    fn add_widget(&mut self, widget_type: widgets::WidgetType,) {
        // Find the next view number we can use
        let next_view_num = self.tree.tiles.next_free_id().0 as usize;

        // Create a new pane with said view number.
        let pane = Pane::with_nr(next_view_num, widget_type.clone());
        
        // Insert it into the tree
        let pane_id = self.tree.tiles.insert_pane(pane);
        // debug!("Inserted pane: {:#?}", pane);


        // Add the new pane to the root tab tile if possible.
        if let Some(root_id) = self.tree.root() {
            // Modify the tile if it's a container and supports children.
            if let Some(root_tile) = self.tree.tiles.get_mut(root_id) {
                if let Tile::Container(container) = root_tile {
                    if let Container::Tabs(tabs) = container {
                        tabs.children.push(pane_id);
                    }
                }
            }
        }
    }
}

// Everything below here I copied directly from the egui_tiles example, I am adapting it. thanks.

/// This is the default state for the egui environment.
impl Default for YamsApp {
    fn default() -> Self {
        // Initialize a variable to keep track of the next view number
        let mut next_view_nr = 0;
        
        // Closure (fancy function?) to create new Panes
        // let mut gen_view = || {
        //     let view = Pane::with_nr(next_view_nr, widgets::WidgetType::About);
        //     next_view_nr += 1;
        //     view
        // };
        // Create new panes for default
        // let about_view = Pane::with_nr(next_view_nr+1, widgets::WidgetType::About);
        let nodelist_view = Pane::with_nr(next_view_nr+1, widgets::WidgetType::NodeList);

        // Create an empty tiles structure to store a tile (which can contain more tiles)
        let mut tiles = egui_tiles::Tiles::default();

        // Vector to hold the ids of our tabs
        let mut tabs = vec![];

        // Insert a single pane tile into our tiles and push its id onto the tabs vector
        // tabs.push(tiles.insert_pane(about_view));
        tabs.push(tiles.insert_pane(nodelist_view));

        // Insert a tab tile with all of our tabs as children into our tiles
        let root = tiles.insert_tab_tile(tabs);

        // Create an egui_tiles::Tree structure to represent our tree of views, starting at the root id
        let tree = egui_tiles::Tree::new("my_tree", root, tiles);

        // Return a new YamsApp instance with our tree and default behavior
        let self_state = Self {
            tree,
            behavior: Default::default(),
        };
        debug!("Full YamsApp default state: {:#?}", &self_state);
        self_state
    }
}

#[cfg_attr(feature = "serde", derive(serde::Deserialize, serde::Serialize))]
/// A struct representing a single pane in the GUI.
/// Each pane has a unique number associated with it.
pub struct Pane {
    /// The unique number of this pane.
    pub nr: usize,
    pub widget_type: widgets::WidgetType,
}

impl std::fmt::Debug for Pane {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        // Write out the debug representation of the Pane struct.
        f.debug_struct("Pane")
            .field("nr", &self.nr) // Include the pane's number in the output.
            .finish()
    }
}

impl Pane {
    /// Creates a new Pane with a unique view number.
    pub fn with_nr(nr: usize, widget_type: widgets::WidgetType) -> Self {
        Self { nr, widget_type }
    }
    /// Draws the UI for the Pane.
    /// This is where we actually put stuff into our panes, like text or buttons.
    /// We also handle dragging of the pane here.
    pub fn ui(&self, ui: &mut egui::Ui) -> egui_tiles::UiResponse {

        // Pass the widget type and ui object down for rendering of the correct type.
        widgets::render_widget(&self.widget_type, ui);
        
        // We don't drag these boys. (surely I can remove this?)
        egui_tiles::UiResponse::None
    }
}

// The storage for the Tree Behaviour config options
#[derive(Debug)]
struct TreeBehavior {
    simplification_options: egui_tiles::SimplificationOptions,
    tab_bar_height: f32,
    gap_width: f32,
    add_child_to: Option<egui_tiles::TileId>,
}

// The defaults for how the tree should behave.
impl Default for TreeBehavior {
    fn default() -> Self {
        Self {
            simplification_options: {
                SimplificationOptions {
                    all_panes_must_have_tabs: true,
                    ..Default::default()
                }
            },
            tab_bar_height: 24.0,
            gap_width: 2.0,
            add_child_to: None,
        }
    }
}


impl TreeBehavior {
    /// The UI for changing TreeBehaviour at runtime
    fn ui(&mut self, ui: &mut egui::Ui) {
        let Self {
            simplification_options,
            tab_bar_height,
            gap_width,
            add_child_to: _,
        } = self;

        // egui::Grid::new("behavior_ui")
        //     .num_columns(2)
        //     .show(ui, |ui| {
        //         ui.label("All panes must have tabs:");
        //         ui.checkbox(&mut simplification_options.all_panes_must_have_tabs, "");
        //         ui.end_row();

        //         ui.label("Join nested containers:");
        //         ui.checkbox(
        //             &mut simplification_options.join_nested_linear_containers,
        //             "",
        //         );
        //         ui.end_row();

        //         ui.label("Tab bar height:");
        //         ui.add(
        //             egui::DragValue::new(tab_bar_height)
        //                 .range(0.0..=100.0)
        //                 .speed(1.0),
        //         );
        //         ui.end_row();

        //         ui.label("Gap width:");
        //         ui.add(egui::DragValue::new(gap_width).range(0.0..=20.0).speed(1.0));
        //         ui.end_row();
        //     });
    }
}


impl egui_tiles::Behavior<Pane> for TreeBehavior {
    fn pane_ui(
        &mut self,
        ui: &mut egui::Ui,
        _tile_id: egui_tiles::TileId,
        view: &mut Pane,
    ) -> egui_tiles::UiResponse {
        view.ui(ui)
    }

    fn tab_title_for_pane(&mut self, view: &Pane) -> egui::WidgetText {
        // Use default Display for most names, but some need to have shit added!
        let mut title = "".to_string();
        match view.widget_type.clone() {
            widgets::WidgetType::NodeInfo => {
                title = format!("Node (SHORTNAME)");
                // We need to get the nodes name.
                // name should be "Node (SHORTNAME)"
                // example: "Node (POCO)"
            },
            widgets::WidgetType::Log => {
                title = format!("{}", widgets::WidgetType::Log)
            },
            widgets::WidgetType::MapView => {
                title = format!("{}", widgets::WidgetType::MapView)
            },
            widgets::WidgetType::Settings => {
                title = format!("{}", widgets::WidgetType::Settings)
            },
            widgets::WidgetType::SignalStrengthHistorical => {
                title = format!("{}", widgets::WidgetType::SignalStrengthHistorical)
            },
            widgets::WidgetType::NodeList => {
                title = format!("{}", widgets::WidgetType::NodeList)
            },
            // why is this the only one not bitching about the widgets:: crap?
            About => {
                title = format!("{}", widgets::WidgetType::About)
            },
        }
        title.into()
    }

    fn top_bar_right_ui(
        &mut self,
        _tiles: &egui_tiles::Tiles<Pane>,
        ui: &mut egui::Ui,
        tile_id: egui_tiles::TileId,
        _tabs: &egui_tiles::Tabs,
        _scroll_offset: &mut f32,
    ) {
        if ui.button("➕").clicked() {
            self.add_child_to = Some(tile_id);
        }
    }

    // ---
    // Settings:

    fn tab_bar_height(&self, _style: &egui::Style) -> f32 {
        self.tab_bar_height
    }

    fn gap_width(&self, _style: &egui::Style) -> f32 {
        self.gap_width
    }

    fn simplification_options(&self) -> egui_tiles::SimplificationOptions {
        self.simplification_options
    }

    fn is_tab_closable(&self, _tiles: &Tiles<Pane>, _tile_id: TileId) -> bool {
        true
    }

    fn on_tab_close(&mut self, tiles: &mut Tiles<Pane>, tile_id: TileId) -> bool {
        if let Some(tile) = tiles.get(tile_id) {
            match tile {
                Tile::Pane(pane) => {
                    // Single pane removal
                    let tab_title = self.tab_title_for_pane(pane);
                    debug!("Closing tab: {}, tile ID: {tile_id:?}", tab_title.text());
                }
                Tile::Container(container) => {
                    // Container removal
                    debug!("Closing container: {:?}", container.kind());
                    let children_ids = container.children();
                    for child_id in children_ids {
                        if let Some(Tile::Pane(pane)) = tiles.get(*child_id) {
                            let tab_title = self.tab_title_for_pane(pane);
                            debug!("Closing tab: {}, tile ID: {tile_id:?}", tab_title.text());
                        }
                    }
                }
            }
        }

        // Proceed to removing the tab
        true
    }
}

#[derive(Debug)]
struct YamsApp {
    tree: egui_tiles::Tree<Pane>,

    #[cfg_attr(feature = "serde", serde(skip))]
    behavior: TreeBehavior,
}

impl eframe::App for YamsApp {
    fn update(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        // Initalise Image Loading (for icons)
        egui_extras::install_image_loaders(ctx);

        // Context/Menu bar at the top
        egui::TopBottomPanel::top("context_bar").show(ctx, |ui| {
            // menu bar sits in here
            menu::bar(ui, |ui| {
                ui.menu_button("File", |ui| {
                    if ui.button("Exit").clicked() {
                        std::process::exit(0);
                    }
                });

                // List of widgets here, we can add new widgets into the panel from here!
                ui.menu_button("Widgets", |ui| {
                    // make a add widget button for every single widget type
                    ui.menu_button("Add Widget", |ui| {
                        if ui.button("About YAMS").clicked() {
                            self.add_widget(widgets::WidgetType::About);
                        }
                        if ui.button("Node List").clicked() {
                            self.add_widget(widgets::WidgetType::NodeList);
                        }
                        if ui.button("Node Info").clicked() {
                            self.add_widget(widgets::WidgetType::NodeInfo);
                        }
                        if ui.button("Logs").clicked() {
                            self.add_widget(widgets::WidgetType::Log);
                        }
                        if ui.button("Map View").clicked() {
                            self.add_widget(widgets::WidgetType::MapView);
                        }
                        if ui.button("Settings").clicked() {
                            self.add_widget(widgets::WidgetType::Settings);
                        }
                        if ui.button("Signal Strength").clicked() {
                            self.add_widget(widgets::WidgetType::SignalStrengthHistorical);
                        }
                    })
                });

                // UI Tools (reset, split? layouts?)
                ui.menu_button("View", |ui| {
                    if ui.button("Reset Panels").clicked() {
                        // Nothing yet
                    }
                });
                // Help
                ui.menu_button("Help", |ui|{
                    if ui.button("Documentation").clicked() {
                        // open git to the wiki / docs page
                    }
                    if ui.button("About").clicked() {
                        // Create a new about widget (don't show it in the widgets menu?)
                        self.add_widget(widgets::WidgetType::About);
                        
                    }
                });
                // Right aligned, logos for showing general status
                // Connection Status enum:
                // ServerNotRunning: mdiLanDisconnect (message: "YAMS is not running")
                // MeshNoConnection: mdiLanPending (message: "YAMS cannot connect to Meshtastic node")
                // MeshConnected: mdiLanConnect (message: "YAMS is connected and receiving data")
                // TODO
                ui.with_layout(egui::Layout::right_to_left(egui::Align::Center), |ui| {
                    ui.add(egui::Image::new(egui::include_image!("../../assets/icons/lan-connect.svg"))
                        .bg_fill(Color32::GREEN)
                        .alt_text("YAMS is connected and receiving data"))
                        // outside the image object
                        .on_hover_text_at_pointer("YAMS is connected and receiving data");
                    if ui.button("Reset").clicked() {
                        *self = Default::default();
                    }
                    self.behavior.ui(ui);
                });
            });
        });

        egui::CentralPanel::default().show(ctx, |ui| {
            self.tree.ui(&mut self.behavior, ui);
        });
    }

    fn save(&mut self, _storage: &mut dyn eframe::Storage) {
        #[cfg(feature = "serde")]
        eframe::set_value(_storage, eframe::APP_KEY, &self);
    }
}

fn tree_ui(
    ui: &mut egui::Ui,
    behavior: &mut dyn egui_tiles::Behavior<Pane>,
    tiles: &mut egui_tiles::Tiles<Pane>,
    tile_id: egui_tiles::TileId,
) {
    // Get the name BEFORE we remove the tile below!
    let text = format!(
        "{} - {tile_id:?}",
        behavior.tab_title_for_tile(tiles, tile_id).text()
    );

    // Temporarily remove the tile to circumvent the borrowchecker
    let Some(mut tile) = tiles.remove(tile_id) else {
        debug!("Missing tile {tile_id:?}");
        return;
    };

    let default_open = true;
    egui::collapsing_header::CollapsingState::load_with_default_open(
        ui.ctx(),
        ui.id().with((tile_id, "tree")),
        default_open,
    )
    .show_header(ui, |ui| {
        ui.label(text);
        let mut visible = tiles.is_visible(tile_id);
        ui.checkbox(&mut visible, "Visible");
        tiles.set_visible(tile_id, visible);
    })
    .body(|ui| match &mut tile {
        egui_tiles::Tile::Pane(_) => {}
        egui_tiles::Tile::Container(container) => {
            let mut kind = container.kind();
            egui::ComboBox::from_label("Kind")
                .selected_text(format!("{kind:?}"))
                .show_ui(ui, |ui| {
                    for typ in egui_tiles::ContainerKind::ALL {
                        ui.selectable_value(&mut kind, typ, format!("{typ:?}"))
                            .clicked();
                    }
                });
            if kind != container.kind() {
                container.set_kind(kind);
            }

            for &child in container.children() {
                tree_ui(ui, behavior, tiles, child);
            }
        }
    });

    // Put the tile back
    tiles.insert(tile_id, tile);
}