use chrono::DateTime;
use chrono_humanize::HumanTime;
use egui::Label;
use egui_extras::{Column, TableBuilder};
use log::info;
use longitude::{DistanceUnit, Location};
use meshtastic::{protobufs::{channel::Role, HardwareModel, Position}, types::NodeId};
use meshtastic::protobufs::{Channel, DeviceMetrics, MeshPacket, MyNodeInfo, NodeInfo, User};

use crate::{db::Database, gui::widgets::node_list::calc_distance_to_node, nodes, CONFIG};

use super::node_list;

// This handles viewing information about a specific node.
// Shows:
// - All the normal info from a node
// - Location History on a map?
// - Recent Messages?
// - Neighbours?
// - Battery Graph
// - Environmental Graph
// - First Seen / Last Heard
// - Distance

/// UI Elements for the Node List
pub fn nodeinfo_ui(ui: &mut egui::Ui) {
    // Get nodes from DB
    let nodes: Vec<NodeInfo> = Database::read_node_list().expect("Failed to read nodes");

    info!("Nodes: {:?}", nodes);

    // Set the first node in the db to be selected
    // This needs to be set elsewhere, but for testing... meh.
    let mut selected_node = nodes.first().unwrap().num;

    // Add a dropdown selector for changing the node
    egui::ComboBox::from_label("Selected Node")
        .selected_text(format!("{:?}", selected_node))
        .show_ui(ui, |ui| {
            for node in &nodes {
                ui.selectable_value(&mut selected_node, node.num, format!("{:?}", node.user.clone().unwrap().id));
            }
        }
    );

    // Iterate through the list until we find the selected node.
    let mut found_selected_node = false;
    for node in &nodes {
        if node.num == selected_node {
            // Render Info when found
            ui.label(format!("Node: {:?}", node));
            ui.label(format!("User ID: {}", node.user.clone().unwrap().id));

            // Compute First Heard
            // How will we do this? the database doesn't store first heard. TODO!

            // Compute Last Heard
            let user = node.user.clone().unwrap();
            let device = node.device_metrics.clone().unwrap();
            // Convert the last heard to a time, then make it nice to read.
            let now = chrono::Local::now().to_utc();
            let humanised = if node.last_heard == 0 {
                String::from("Unknown")
            } else {
                let last_heard = DateTime::from_timestamp(node.last_heard.into(), 0).unwrap();
                let duration = last_heard - now;
                HumanTime::from(duration).to_string()
            };
            ui.label(format!("Last Heard: {}", humanised));

            // Find Messages to or from said node
            // TODO: Requires parsing all MeshPackets in the DB
            
            // Find all Battery stats from said node
            // TODO: Requires parsing all MeshPackets in the DB

            // Find all temperature stats from said node
            // TODO: Requires parsing all MeshPackets in the DB

            // Find Location history from said node
            // TODO: Requires parsing all MeshPackets in the DB
            // Perhaps with the 3 above, we parse all packets to or from the node, and then filter depending on the above 4

            // Find distance to node.
            if let Some(distance) = calc_distance_to_node(node.position.clone().unwrap()) {
                if distance.is_nan() {
                    ui.label("-");
                } else {
                    // Get the unit shorthand from the GuiConfig struct
                    let unit_shorthand = CONFIG.gui.get_unit_shorthand().unwrap_or("");
                    ui.label(format!("Distance to Node: {:.2}{}", distance, unit_shorthand));
                }
            } else {
                ui.label("-");
            }

            // Find Direct (radio) Neighbours of said node
            
            found_selected_node = true;
            break;  // Exit loop once the selected node is found
        }
    }

    if !found_selected_node {
        ui.label("Selected node not found.");
    }

}