use std::{env, process};

use figment::{Figment, providers::{Format, Toml, Env}};

use lazy_static::lazy_static;
use meshtastic::api::StreamApi;
use meshtastic::utils;

use packets::parse;
use log::{error, warn, info, debug, trace};

mod nodes;
mod packets;
mod db;
mod config;
mod gui;

use crate::config::AppConfig;

// Setup the config globally because I can't figure out how to pass it to functions I don't call directly.
// This is evaluated at runtime, and not compilation. \o/
// Also, we try to merge configs. so envvars > config > defaults > error
lazy_static! {
    pub static ref CONFIG: AppConfig = {
        let config = Figment::new()
            .merge(Toml::file("config.toml"))
            .merge(Env::prefixed("yamm_"))
            .extract()
            .unwrap_or_default();
        info!("Loaded Config!");
        debug!("full config: {:?}", config);
        config
    };
}

async fn run_server() {
    // Start Logging
    colog::init();

    // set up new connection
    let stream_api = StreamApi::new();

    // Different connection types have different setups.
    match &CONFIG.meshconfig.connection_type {
        config::ConnectionType::Tcp => {
            // Build the tcp stream object from config.
            let address = &CONFIG.meshconfig.ip;
            let port = &CONFIG.meshconfig.port.to_string();
            let tcp = format!("{}:{}", address, port);

            // Connect over TCP
            let tcp_stream = utils::stream::build_tcp_stream(tcp).await.expect("Invalid TCP Stream");
            let (mut decoded_listener, stream_api) = stream_api.connect(tcp_stream).await;
            let config_id = utils::generate_rand_id();
            let stream_api = stream_api.configure(config_id).await.expect("Invalid Stream Config");

            // Read loop, exit with Ctrl+C or disconnecting node.
            while let Some(decoded) = decoded_listener.recv().await {
                // New Message!
                debug!("Received: {:?}", decoded);
                // Parse message
                parse(&decoded);
            }
        
            // Note that in this specific example, this will only be called when
            // the radio is disconnected, as the above loop will never exit.
            // Typically you would allow the user to manually kill the loop,
            // for example with tokio::select!.
            let _stream_api = stream_api.disconnect().await.expect("Failed to disconnect stream");
        },
        config::ConnectionType::Serial => {
            let available_ports = utils::stream::available_serial_ports().expect("Couldn't find serial ports");
            info!("Available ports, please set in config.: {:?}", available_ports);
            todo!("Not yet Implemented. See https://git.volkor.me/Volkor/yamm/-/issues/2")
        },
        config::ConnectionType::Mqtt => {
            todo!("Not yet Implemented. See https://git.volkor.me/Volkor/yamm/-/issues/3")
        },
    }
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    // Do we run a server or gui?
    let args: Vec<String> = env::args().collect();

    // display help on no args, exit.
    if args.len() < 2 {
        println!("Usage: yams <mode>");
        println!("Modes: 'server', 'gui'");
        process::exit(1);
    }

    // match on gui or server
    match args[1].as_str() {
        "server" => run_server().await,
        "gui" => gui::run_gui().await.expect("uh oh"),
        _ => {
            eprintln!("Unknown mode: {}", args[1]);
            println!("Usage: yams <mode>");
            println!("Modes: 'server', 'gui'");
            process::exit(1);
        }
    }

    Ok(())
}

pub mod built_info {
    include!(concat!(env!("OUT_DIR"), "/built.rs"));
}